# js dec2024 cleaned up code with vulture,
#    updated library calls for OSMnx version 2.0.0
#    updated library calls for Folium 0.19.2

import osmnx as ox
import networkx as nx
import folium
import pickle
import time, datetime
import os.path
import sys
import webbrowser

import os

import matplotlib.pyplot as plt
plt.switch_backend('agg')


ox.settings.log_console=True
ox.settings.use_cache=True

print('OSMnx version: ', ox.__version__)
print('NetworkX version: ',nx.__version__)
print('Python version: ',sys.version)

start_time = time.time()

def elapsed_time(start):
    end_time = time.time()
    duration = round(end_time - start)
    timestamp = str(datetime.timedelta(seconds = duration))
    print(timestamp)


def report(Q, txt=''):
    print(len(Q.nodes()), ' nodes ', txt)
    print(len(Q.edges()), ' edges ', txt)
    if isinstance(Q,nx.MultiGraph):    
        length = ox.stats.edge_length_total(Q) #in meters
        length = length * 0.0006213712
        print(f'total length {length:.2f} miles ', txt)
    zeroes = 0    
    odds = 0
    evens = 0
    twos = 0
    ones = 0
    fours = 0
    four_pairs = 0
    parallels = 0
    for node in Q.nodes:
        d = Q.degree(node)
        if d == 0:
            zeroes += 1
        if d % 2 == 1:
            odds += 1
        else:
            evens += 1
        if d == 2:
            twos += 1
        if d == 1:
            ones += 1
        if d==4:
            fours += 1
            for neighbor in Q.neighbors(node):
                if Q.degree(neighbor)==4 and neighbor>node: # prevent double-count
                    four_pairs += 1
        if d>6:
            print('node ', node, 'has degree ', d)
        for neighbor in Q.neighbors(node):
            edge_count = Q.number_of_edges(node,neighbor)
            if edge_count > 1:
                parallels += edge_count - 1
                
    if zeroes > 0:            
        print('number of zero-degree nodes ' + txt + ' :', zeroes)            
    print('number of odd intersections ' + txt + ' :', odds)
    print('number of even intersections ' + txt + ' :', evens)
    print('number of 2-way intersections ' + txt + ' :', twos)
    print('number of 4-way intersections ' + txt + ' :', fours)
    print('number of trailhead/terminus ' + txt + ' :', ones)
    print('number of parallel edges ' + txt + ' :', parallels)
    if four_pairs > 0:
        print('number of 4-way adjacent pairs ' + txt + ' :', four_pairs)

        
    check_node_geometry(Q)
    print('')

def colorize_graph(Q):
    QR = Q.copy()
    for node, dat in Q.nodes(data=True):
        d = Q.degree(node)
        dat['color'] = d

        QR.add_node(node,**dat)                #replace node value

    for u, v, k, dat in Q.edges(keys=True,data=True):
        d = 'path'
        if 'highway' in dat:
            d = dat['highway']
        dat['color']=0.0
        if isinstance(d, str):
            if d=='path' or d == 'footway':
                dat['color']=1.0
            else:
                dat['color'] = 2.0
        QR.add_edge(u,v,key=k,**dat)
    QR.graph['color']=True
    return QR

                 
def check_node_geometry(Q):
    count = 0
    for node, dat in Q.nodes(data=True):
        if 'x' not in dat:
            count += 1
            #d = Q.degree(node)
            #print('node ', node, ' has no geometry attribute, degree ', d)t
    if count>0:
        print('number of nodes without geometry: ',count)


def check(Q,Q_original):
    remainder_nodes = get_remainder_odd_nodes(Q)
    #print('number of odd nodes remaining: ', len(remainder_nodes))
    check_remainder_odd_nodes(Q_original,remainder_nodes)
    for u,v,k,dic in Q.edges(data=True, keys=True):
        if 'node4_path' in dic:
            print('temporary path not deleted between: ',u,' and ',v, ' with key ', k)
        if not 'highway' in dic:
            print('suspicious edge: ',u,' and ',v, ' with key ', k)


def get_remainder_odd_nodes(Q):
    node_list = []
    for node in Q.nodes:
        if Q.degree(node) %2 ==1:
            node_list.append(node)
    return node_list

# remove trailheads and termini
def delete_ends(Q):
    while True: # simulate do-while
        rn = []
        for node, neighbors in Q.adjacency():
            if len(neighbors) == 1:
                rn.append(node)
        if len(rn)>0:            
            Q.remove_nodes_from(rn)
        else:
            break

def check_remainder_odd_nodes(Q_original,node_list):
    remainder_odd_node_count = 0
    remainder_odd_node_4node_neighber_count = 0
    for node in node_list:
        if Q_original.degree(node)==4:
            remainder_odd_node_count +=1
            #print('remainder odd node ', node, 'was a 4node')
        else:
            for neighbor in Q_original.neighbors(node):
                if Q_original.degree(neighbor)==4:
                    #print('remainder odd node ',node,' has neighbor ', neighbor,' that is a 4node')
                    remainder_odd_node_4node_neighber_count +=1
    if remainder_odd_node_count > 0:
        print('remainder odd_node count: ', remainder_odd_node_count)
    if remainder_odd_node_4node_neighber_count > 0:
        print('remainder odd node with 4node neighbor count: ', remainder_odd_node_4node_neighber_count)
        

     
town_markers = [(34.29963, -108.13268, 'Pie Town'),
         (33.19593, -108.20834, 'Gila Hot Springs'),
         (32.77981, -108.27533, 'Silver City'),
         (33.37939, -108.90369, 'Alma/Glenwood'),
         (33.03022, -108.16864, 'Lake Roberts'),
         (32.85594, -107.98024, 'Mimbres'),       
         (32.91685, -107.70593, 'Kingston')]

# draw() changed after moving to Linux
def draw(Q,title='',show_nodes = False, boxes=None, not_boxes = None, markers=town_markers):
    start_latlng = (33.04876, -108.30680)
    m = folium.Map(location=[start_latlng[1],start_latlng[0]],
               zoom_start=12)
    # convert to MultiDiGraph for folium
    T = Q.copy()
    if not 'color' in T.graph:
        T = colorize_graph(T)
    QD = T.to_directed()
    QD.graph['crs'] = ox.settings.default_crs

    # our graphs may be HUGE, so simplify if not already done so
    if not "simplified" in QD.graph:
        QD = ox.simplify_graph(QD, remove_rings=True, track_merged=True)
    
    nodes, edges = ox.graph_to_gdfs(QD)
#    m = edges.explore(color='#ff00ff', tiles="openstreetmap",tooltip='name')
    m = edges.explore(column='color',  cmap = 'tab10',legend=False, tiles="openstreetmap",tooltip='name')
    if show_nodes:    
#        nodes.explore(m=m, color="pink", marker_kwds={"radius": 4})    
        nodes.explore(m=m, column='color', cmap='tab10', legend=False,marker_kwds={"radius": 4})    

    if not boxes == None:
        for n,s,e,w in boxes:
            folium.Rectangle(bounds=[(n,e),(n,w),(s,e),(s,w)], color='#ff7800', fill=True, fill_color='#ffff00', fill_opacity=0.2).add_to(m)
    if not not_boxes == None:
        for n,s,e,w in not_boxes:
            folium.Rectangle(bounds=[(n,e),(n,w),(s,e),(s,w)], color='#00ff78', fill=False, fill_color='#00ffff', fill_opacity=0.2).add_to(m)
    if not markers == None:
        for lat,lon, txt in markers:
            #print(lat,lon,txt)
            folium.Marker(location=[lat,lon],popup=txt,
                          icon=folium.Icon(color='lightred',icon=''),
                          tooltip=txt).add_to(m)
            
    # include the option to switch map layers
   # folium.TileLayer('openstreetmap').add_to(m)
    folium.TileLayer('cartodbpositron').add_to(m)
   # folium.TileLayer('Stamen Terrain').add_to(m)
    folium.TileLayer(
        tiles='https://{s}.tile-cyclosm.openstreetmap.fr/cyclosm/{z}/{x}/{y}.png',
        name='cyclosm',
        attr='CyclOSM',
    ).add_to(m)
    folium.LayerControl().add_to(m)    

    title_html = '''
            <h3 align="center" style="font-size:16px"><b>{}</b></h3>
             '''.format(title)   
    m.get_root().html.add_child(folium.Element(title_html))
    file_name = "/home/sagebrush/biz/graph/map_" + title + ".html"
    m.save(file_name)
    #m.show_in_browser()
    webbrowser.open(file_name)
    input("Press Enter to continue...")
    print('*************', flush=True)
    print("", flush=True)    
        


J = pickle.load(open('j_maximum_with_roadwalks.pkl', 'rb'))
K = pickle.load(open('maxt2_with_roadwalks_788.pkl', 'rb'))
report(K,'candidate graph ')
draw(K,'max_candidate_788', show_nodes=True)

add_edges = [
##            (142198578,533600800,'silver n patch'),
##                    (5336008002, 142198743, 'silver ne patch'),
                (2510834192,5804066152, 'silver n to e'),
              (5804066152,142198578, 'silver e to ne'),
              (5804066152,6226324419, 'silver e to ne'),
              

              ]

            
remove_edges = [ 
                (2510834192 , 142198578), #silver bypass
                   ]

                 
for u, v in remove_edges:
    try:
        K.remove_edge(u,v)
        J.remove_edge(u,v) #somewhat dangerous
    except Exception as ex:
        print('no path found: ', u,v)

not_recommended = ['sawmill wagon road trail', #179
                   '*bear canyon road',
                   ]
edges_to_remove = []
trails_removed = set()
for u,v,k,d in K.edges(keys=True,data=True):
    if 'name' in d:
        trail_name = d['name']
        if type(trail_name) is list:
            trail_name = ' '.join(trail_name)
        trail_name = trail_name.lower()
#        print(trail_name)
        for nr in not_recommended:
            if nr in trail_name or (nr.startswith('*') and nr.strip('*')==trail_name):
#        if any(e for e in not_recommended if e in trail_name):
                edges_to_remove.append( (u,v,k) )
                trails_removed.add(trail_name)
                print('removed: ', trail_name, ' from ', nr)
K.remove_edges_from(edges_to_remove)
#J.remove_edges_from(edges_to_remove)

#K.add_edge(142232057,   142232076)



#report(J,'candidate with added edges')

# fix: this crashes
#draw(J,'simplified graph',show_nodes=True)
#draw(K,'candidate 786',show_nodes=True)

G2 = nx.MultiGraph()
for node,node2,description in add_edges:
    try:
        if description.startswith('*'):
            path = nx.shortest_path(J,node,node2)
        else:
            path = nx.shortest_path(J,node,node2,weight='length')
        path_length = nx.path_weight(J,path,weight='length')
    except Exception as ex:
        print('no path found: ', description, ex, 'nodes: ',node,node2)
    else:
        print(description,' distance(km): ', round(path_length/1000,1),
              'nodes: ', node,node2, ' description: ', description)
        if path != None and len(path)>0:
            T = J.subgraph(path).copy()
            G2 = nx.compose(G2,T)
        
K = nx.compose(K,nx.MultiGraph(nx.Graph(G2))) #sneaky way to avoid parallel edge

# a little cleanup if needed
delete_ends(K)
nodes = max(nx.connected_components(K), key=len) #remove any isolated islands
K = nx.subgraph(K, nodes).copy()
check(K,J)
       
if nx.is_eulerian(K):
    print('k is eulerian!!!!!')
else:
    print('back to the drawing board')
    K = colorize_graph(K)
    draw(K,'back to drawing board',show_nodes=True)

            
               

check(J,J)



def fix_node_geometry(Q,Q_original):
    #restore_count = 0
    QR = Q.copy()
    for node in Q.nodes():  # add node attributes back in
        node_dic = Q.nodes[node]
        if 'x' not in node_dic:
            node_dic = Q_original.nodes[node]
            QR.add_node(node,**node_dic)
            #restore_count += 1
   # if restore_count > 0:
    #    print('number of 4nodes restored: ', restore_count)
    return QR
    
K = fix_node_geometry(K,J)       
    
print('finished adding towns')


pickle.dump(K,open('max_candidate_with_towns.pkl', 'wb'))
draw(K,'max_candidate_with_towns')


elapsed_time(start_time)

print('finished script ' + os.path.basename(__file__))






